from strategy_api.tm_api.object import Exchange, Product, OrderType


def symbol_deal(symbol: str, exchange: Exchange) -> str:
    if "-" not in symbol:
        print("错误的标的，主币 与 订价币之间用 - ")
        return symbol

    l = symbol.split('-')
    if exchange == Exchange.BINANCE:
        if "PERP" in l:
            return l[0] + l[1] + "_PERP"
        else:
            return l[0] + l[1]
    elif exchange == Exchange.OKEX:
        if "PERP" in l:
            return symbol.replace("PERP", "SWAP")
        else:
            return symbol + "-SWAP"
    else:
        print("未知交易所")
        return ""


def get_order_type(maker: bool, stop_loss: bool, stop_profit: bool, exchange: Exchange) -> OrderType:
    if maker:
        # 发送限价单
        order_type = OrderType.LIMIT
    else:
        if stop_profit and stop_loss:
            if exchange == Exchange.BINANCE:
                raise ValueError("BINANCE cannot combine stop loss and take profit orders")
            order_type = OrderType.STOP_LOSS_PROFIT
            return order_type
        if stop_profit:
            # 发送市价止盈
            order_type = OrderType.TAKE_PROFIT_MARKET
        elif stop_loss:
            # 发送市价止损
            order_type = OrderType.STOP_MARKET
        else:
            # 发送普通市价单
            order_type = OrderType.MARKET
    return order_type



# tick 数据: 最后的市场交易、订单快照、盘中市场统计
from dataclasses import dataclass
from datetime import datetime

# tick 数据处理
from enum import Enum
from types import TracebackType
from typing import Union, Callable, Type, Any

CALLBACK_TYPE = Callable[[Union[dict, list], "Request"], None]
ON_FAILED_TYPE = Callable[[int, "Request"], None]
ON_ERROR_TYPE = Callable[[Type, Exception, TracebackType, "Request"], None]


# ---------------------------------枚举-----------------------------
class DataType(Enum):
    TICK = "tick"
    BAR = "bar"


class Interval(Enum):
    """
    Interval of bar data.

    """
    MINUTE = "1m"
    MINUTE_3 = "3m"
    MINUTE_5 = "5m"
    MINUTE_15 = "15m"
    MINUTE_30 = "30m"

    HOUR = "1h"
    HOUR_2 = "2h"
    HOUR_4 = "4h"
    HOUR_6 = "6h"
    HOUR_8 = "8h"
    HOUR_12 = "12h"


# 订单状态类型
class Status(Enum):
    """
    Order status.
    """
    SUBMITTING = "SUBMITTING"  # 提交中 status
    NOTTRADED = "NOTTRADED"  # 没有交易 status
    PARTTRADED = "PARTTRADED"  # 部分交易 status
    ALLTRADED = "ALLTRADED"  # 全部交易 status
    CANCELLED = "CANCELLED"  # 撤单 status
    REJECTED = "REJECTED"  # 拒单 status


ACTIVE_STATUSES = set([Status.SUBMITTING, Status.NOTTRADED, Status.PARTTRADED])


# 订单、交易、仓位等的 方向描述
class Direction(Enum):
    LONG = "Long"  # 多
    SHORT = "Short"  # 空


# 订单类型
class OrderType(Enum):
    """
    Order type.
    """
    LIMIT = "LIMIT"  # 限价单
    MARKET = "MARKET"  # 市价单
    STOP_MARKET = "STOP_MARKET"  # 市价止损
    TAKE_PROFIT_MARKET = "TAKE_PROFIT_MARKET"  # 世价止盈
    STOP_LOSS_PROFIT = "STOP_LOSS_PROFIT"  # 止盈止损


class PositionSide(Enum):
    ONEWAY = "ONE_WAY"
    TWOWAY = "TWO_WAY"


# 清算：指订单、成交中的描述
class Offset(Enum):
    NONE = ""
    OPEN = "OPEN"  # 开
    CLOSE = "CLOSE"  # 关


# 产品类型
class Product(Enum):
    SPOT = "SPOT"  # 现货
    U_FUTURES = "U_FUTURES"  # 合约
    B_FUTURES = "B_FUTURES"  # 合约


# 交易所名称
class Exchange(Enum):
    BINANCE = "BINANCE"
    OKEX = "OKEX"


# -----------------------------------------------------------

@dataclass
class BarData:
    symbol: str
    datetime: datetime
    endTime: datetime

    interval: Interval = None
    volume: float = 0
    turnover: float = 0
    open_price: float = 0
    high_price: float = 0
    low_price: float = 0
    close_price: float = 0
    exchange: Exchange = None  # 交易所
    product: Product = None


@dataclass
class TickData:
    symbol: str
    datetime: datetime

    volume: float = 0
    turnover: float = 0
    last_price: float = 0
    last_volume: float = 0
    limit_up: float = 0
    limit_down: float = 0

    open_price: float = 0
    high_price: float = 0
    low_price: float = 0
    pre_close: float = 0
    localtime: datetime = None
    exchange: Exchange = None  # 交易所
    product: Product = None


# order 数据：当前挂单
@dataclass
class OrderData:
    symbol: str
    orderid: str

    type: OrderType = OrderType.LIMIT
    direction: Direction = None
    offset: Offset = Offset.NONE
    price: float = 0
    volume: float = 0
    traded: float = 0
    traded_price: float = 0
    status: Status = Status.SUBMITTING
    datetime: datetime = datetime.now()
    update_time: datetime = datetime.now()
    reference: str = ""
    rejected_reason: str = ""  # Order Rejected Reason
    exchange: Exchange = None  # 交易所

    # 检查订单是否有效。
    def is_active(self) -> bool:
        return self.status in ACTIVE_STATUSES

    # 创建 一个测序订单请求数据
    def create_cancel_request(self) -> "CancelRequest":
        return CancelRequest(
            orderid=self.orderid, symbol=self.symbol
        )


# 订单请求数据 发送到特定网关以创建新订单
@dataclass
class OrderRequest:
    """
    Request sending to specific gateway for creating a new order.
    """

    symbol: str
    direction: Direction  # 做多做空方向
    type: OrderType  # 订单类型： 限价、市价、市价止盈、止损
    volume: float  # 交易量
    price: float = 0  # 价格
    offset: Offset = Offset.NONE  # 抵消
    reference: str = ""  # 参考
    positionSide: PositionSide = PositionSide.ONEWAY
    exchange: Exchange = None  # 交易所

    stop_loss_price: float = 0  # 止损触发价格
    stop_profit_price: float = 0  # 止盈触发价

    def create_order_data(self, orderid: str) -> OrderData:
        order: OrderData = OrderData(
            symbol=self.symbol,
            orderid=orderid,
            type=self.type,
            direction=self.direction,
            offset=self.offset,
            price=self.price,
            volume=self.volume,
            reference=self.reference,
            exchange=self.exchange
        )
        return order


@dataclass
class CancelRequest:
    """
    """

    orderid: str
    symbol: str
    exchange: Exchange = None  # 交易所


@dataclass
class TradeData:
    symbol: str
    orderid: str
    tradeid: str = ""
    direction: Direction = None

    offset: Offset = Offset.NONE
    price: float = 0
    volume: float = 0
    datetime: datetime = None
    exchange: Exchange = None  # 交易所


@dataclass
class HistoryRequest:
    symbol: str
    start: datetime
    end: datetime = None
    interval: Interval = None
    exchange: Exchange = None  # 交易所


@dataclass
class ContractData:
    symbol: str  # 简称
    exchange: Exchange  # 交易所
    name: str  # 全称
    product: Product  # 类型
    size: float  # 合约乘数
    pricetick: float  # 下单价格精度
    min_volume: float = float(1)  # 最小下单数量



# 导入 strategy_api 包里面的策略模版
from strategy_api.strategies.template import StrategyTemplate
# 导入 BINANCE U 本位合约 API包
from strategy_api.tm_api.Okex.futureUsdt import OkexFutureUsdtGateway
from strategy_api.tm_api.Binance.futureInverse import BinanceFutureInverseGateway

# 导入 strategy_api 中使用的常量对象
from strategy_api.tm_api.object import Interval, BarData, OrderData, Status, PositionSide, DataType, Exchange


# 策略类
class StrategyDemo(StrategyTemplate):
    # 属性 作者(标志该策略的开发人员)
    author = "DYX"

    binance_api_setting = {
        "key": "",
        "secret": "",
        "proxy_host": "127.0.0.1",
        "proxy_port": 8010,
        "Passphrase": "",

        "call_order_switch": False,  # 订单数据回调 开关
        "call_tick_switch": False,  # tick数据回调 开关
        "call_bar_switch": True  # k线数据回调 开关
    }

    okex_api_setting = {
        "key": "",
        "secret": "",
        "proxy_host": "127.0.0.1",
        "proxy_port": 8010,
        "Passphrase": "",

        "call_order_switch": True,  # 订单数据回调 开关
        "call_tick_switch": False,  # tick数据回调 开关
        "call_bar_switch": True  # k线数据回调 开关
    }

    # 初始化方法
    def __init__(self):
        super(StrategyDemo, self).__init__()

    # 初始化策略参数
    def init_parameters(self):
        self.symbol = "SOL-USD-PERP"

        self.buy_switch = True
        self.long_id = ""
        self.stop_loss_id = ""
        self.stop_profit_id = ""
        self.volume = 1
        self.rate_stop = 0.008

        # 添加 BINANCE U本位网关
        okex_gateway = self.add_gateway(OkexFutureUsdtGateway, "okex_api", self.okex_api_setting)
        binance_gateway = self.add_gateway(BinanceFutureInverseGateway, "binance_api", self.binance_api_setting)

        # 订阅数据
        okex_gateway.subscribe(symbol=self.symbol, data_type=DataType.BAR, interval=Interval.MINUTE)
        binance_gateway.subscribe(symbol=self.symbol, data_type=DataType.BAR, interval=Interval.MINUTE)

    # k 线数据的回调, 可以在该方法里面记录 k 线数据、分析k线数据
    def on_bar(self, bar: BarData):
        okex_api = self.get_gateway("okex_api")

        print(bar)

        if self.buy_switch and bar.exchange == Exchange.OKEX:
            # 开多单
            self.long_id = okex_api.new_order_id()

            okex_api.buy(
                orderid=self.long_id,
                symbol=self.symbol,
                volume=self.volume,
                price=10,
                position_side=PositionSide.TWOWAY
            )
            print("--------------做多")
            self.buy_switch = False

    # 获取历史k线，获取最新一根k线的开盘价
    def query_history_kline(self):
        okex_api = self.get_gateway("okex_api")
        try:
            kls = okex_api.query_history(symbol=self.symbol, minutes=1, interval=Interval.MINUTE)
        except Exception:
            kls = okex_api.query_history(symbol=self.symbol, minutes=1, interval=Interval.MINUTE)

        open_price = kls[-1].open_price
        return open_price

    # 订单 数据的回调，订单状态的改变都会通过websoket 推送到这里，例如 从提交状态 改为 全成交状态，或者提交状态 改为 撤销状态 都会推送
    # 可以在这里对仓位进行一个记录
    def on_order(self, order: OrderData):
        okex_api = self.get_gateway("okex_api")
        print(order)
        if order.status == Status.ALLTRADED and self.long_id == order.orderid:
            print("做多成交")
            open_price = self.query_history_kline()
            print(f"最新开盘价：{open_price}")
            self.stop_id = okex_api.new_order_id()
            okex_api.sell(
                orderid=self.stop_profit_id,
                symbol=self.symbol,
                volume=self.volume,
                price=round(open_price * (1 + self.rate_stop), 3),
                stop_profit=True,
                stop_profit_price=round(open_price * (1 + self.rate_stop), 3),
                stop_loss=True,
                stop_loss_price=round(open_price * (1 - self.rate_stop), 3),
                position_side=PositionSide.TWOWAY
            )

        elif order.status == Status.ALLTRADED and self.stop_id == order.orderid:
            # okex 不用测小止损止盈单，只要止损了，止盈单自动撤销
            print("止盈或者止损了")
            print(order)
            self.buy_switch = True


def start_strategy(api_setting):
    # 初始化策略
    s = StrategyDemo()




if __name__ == '__main__':
    print("启动量化系统: 等待策略运行")
    api_setting = {
        "key": "",
        "secret": "",
        "proxy_host": "",
        "proxy_port": 0,
        "Passphrase": "",

        "call_order_switch": False,  # 订单数据回调 开关
        "call_tick_switch": False,  # tick数据回调 开关
        "call_bar_switch": False  # k线数据回调 开关
    }
    start_strategy(api_setting)



from abc import ABC
from typing import Dict, Any, List
from strategy_api.event.engine import EVENT_TICK, EVENT_BAR, EVENT_ORDER
from strategy_api.event.engine import EventEngine, Event
from strategy_api.tm_api.object import OrderData, BarData, TickData, Interval, OrderRequest, HistoryRequest, \
    PositionSide, DataType, Exchange, Product


class BaseGateway(ABC):
    # Default name for the gateway.
    default_setting: Dict[str, Any] = {}
    exchange: Exchange = None
    product: Product = None
    symbolMap: dict = dict()

    def __init__(self,
                 event_engine: EventEngine,
                 order_switch: bool = False,
                 bar_switch: bool = False,
                 tick_switch: bool = False) -> None:

        self.event_engine: EventEngine = event_engine

        self.order_call_switch = order_switch
        self.bar_call_switch = bar_switch
        self.tick_call_switch = tick_switch

    def connect(self, setting: dict) -> None:
        pass

    def buy(self,
            orderid: str,
            symbol: str,
            volume: float,  # 数量
            price: float = 0,  # 价格
            maker: bool = False,  # 限价单
            stop_loss: bool = False,  # 止损
            stop_loss_price: float = 0,  # 止损价
            stop_profit: bool = False,  # 止盈
            stop_profit_price: float = 0,  # 止盈价
            position_side: PositionSide = PositionSide.ONEWAY,
            ):
        pass

    def sell(self,
             orderid: str,
             symbol: str,
             volume: float,  # 数量
             price: float = 0,  # 价格
             maker: bool = False,  # 限价单
             stop_loss: bool = False,  # 止损
             stop_loss_price: float = 0,  # 止损价
             stop_profit: bool = False,  # 止盈
             stop_profit_price: float = 0,  # 止盈价
             position_side: PositionSide = PositionSide.ONEWAY,
             ):
        pass

    def short(self,
              orderid: str,
              symbol: str,
              volume: float,  # 数量
              price: float = 0,  # 价格
              maker: bool = False,  # 限价单
              stop_loss: bool = False,  # 止损
              stop_loss_price: float = 0,  # 止损价
              stop_profit: bool = False,  # 止盈
              stop_profit_price: float = 0,  # 止盈价
              position_side: PositionSide = PositionSide.ONEWAY,
              ):
        pass

    def cover(self,
              orderid: str,
              symbol: str,
              volume: float,  # 数量
              price: float = 0,  # 价格
              maker: bool = False,  # 限价单
              stop_loss: bool = False,  # 止损
              stop_loss_price: float = 0,  # 止损价
              stop_profit: bool = False,  # 止盈
              stop_profit_price: float = 0,  # 止盈价
              position_side: PositionSide = PositionSide.ONEWAY,
              ):
        pass

    def new_order_id(self) -> str:
        pass

    def send_order(self, req: OrderRequest, orderid: str):
        pass

    def cancel_order(self, orderid: str, symbol: str) -> None:
        pass

    def query_history(self, symbol: str, interval: Interval, hour: int = 0, minutes: int = 0) -> List[BarData]:
        pass

    def subscribe(self, symbol: str, data_type: DataType, interval: Interval = None) -> None:
        pass

    # 一般事件推送。
    def on_event(self, type: str, data: Any = None) -> None:
        event: Event = Event(type, data)
        self.event_engine.put(event)

    # tick 数据推送
    def on_tick(self, tick: TickData) -> None:
        if self.tick_call_switch:
            tick.symbol = self.symbolMap[tick.symbol]
            self.on_event(EVENT_TICK, tick)

    # K线数据回调
    def on_bar(self, bar: BarData):
        if self.bar_call_switch:
            bar.symbol = self.symbolMap[bar.symbol]
            self.on_event(EVENT_BAR, bar)

    # 定单事件推送
    def on_order(self, order: OrderData) -> None:
        if self.order_call_switch:
            order.symbol = self.symbolMap[order.symbol]
            self.on_event(EVENT_ORDER, order)

    def close(self) -> None:
        pass



# -*- coding:utf-8 -*-
# @Date: 2022-11-01
# @author: 邓大大
# @Desc: winTrader 框架的事件驱动框架。
# @Notice: 事件引擎开始后, 默认启动了一个 计时器
from collections import defaultdict
from queue import Empty, Queue
from threading import Thread
import time
from typing import Any, Callable, List

EVENT_TIMER = "eTimer"
EVENT_TICK = "eTick"
EVENT_BAR = "eBar"
EVENT_ORDER = "eOrder"


# 事件对象
# 由使用的类型字符串组成,通过事件引擎分发事件，以及一个数据
class Event:

    def __init__(self, type: str, data: Any = None) -> None:
        """"""
        self.type: str = type
        self.data: Any = data


# 定义要在事件引擎中使用的处理函数。
HandlerType: callable = Callable[[Event], None]


# 事件引擎
# 1、根据其类型分发事件对象给那些注册的处理程序。
# 2、它还按间隔秒生成计时器事件，可用于计时目的： 定时器事件默认每 1 秒生成一次，如果未指定间隔。
class EventEngine:

    def __init__(self, interval: int = 1) -> None:
        self._interval: int = interval
        self._queue_order: Queue = Queue()
        self._queue_tick: Queue = Queue()
        self._queue_bar: Queue = Queue()
        self._queue_time: Queue = Queue()
        self._active: bool = False
        self._timer: Thread = Thread(target=self._run_timer)

        self._timer_er: Thread = Thread(target=self._run_timer_er)
        self._barer: Thread = Thread(target=self._run_bar)
        self._ticker: Thread = Thread(target=self._run_tick)
        self._order_er: Thread = Thread(target=self._run_order)

        self._order_handlers: defaultdict = defaultdict(list)  # 特殊方法对应不同事件类型
        self._bar_handlers: defaultdict = defaultdict(list)  # 特殊方法对应不同事件类型
        self._tick_handlers: defaultdict = defaultdict(list)  # 特殊方法对应不同事件类型
        self._time_handlers: defaultdict = defaultdict(list)  # 特殊方法对应不同事件类型

    def _run_timer_er(self) -> None:
        while self._active:
            try:
                event: Event = self._queue_time.get(block=True, timeout=1)
                # 注册的每个方法都一次处理该对象
                [handler(event) for handler in self._time_handlers[event.type]]
            except Empty:
                pass

    def _run_order(self) -> None:
        while self._active:
            try:
                event: Event = self._queue_order.get(block=True, timeout=1)
                # 注册的每个方法都一次处理该对象
                [handler(event) for handler in self._order_handlers[event.type]]
            except Empty:
                pass

    def _run_bar(self) -> None:
        while self._active:
            try:
                event: Event = self._queue_bar.get(block=True, timeout=1)
                # 注册的每个方法都一次处理该对象
                [handler(event) for handler in self._bar_handlers[event.type]]
            except Empty:
                pass

    def _run_tick(self) -> None:
        while self._active:
            try:
                event: Event = self._queue_tick.get(block=True, timeout=1)
                # 注册的每个方法都一次处理该对象
                [handler(event) for handler in self._tick_handlers[event.type]]
            except Empty:
                pass

    # 定时事件put
    def _run_timer(self) -> None:
        while self._active:
            time.sleep(self._interval)
            event: Event = Event(EVENT_TIMER)
            self.put(event)

    # 开始事件引擎
    def start(self):
        self._active = True
        self._timer.start()

        self._ticker.start()
        self._barer.start()
        self._order_er.start()
        self._timer_er.start()

    # 停止事件引擎
    def stop(self):
        self._active = False
        self._timer.join()

        self._ticker.join()
        self._barer.join()
        self._order_er.join()
        self._timer_er.join()

    # 事件对象加入到事件队列中待处理
    def put(self, event: Event) -> None:
        if event.type == EVENT_TIMER:
            self._queue_time.put(event)
        elif event.type == EVENT_ORDER:
            self._queue_order.put(event)
        elif event.type == EVENT_TICK:
            self._queue_tick.put(event)
        elif event.type == EVENT_BAR:
            self._queue_bar.put(event)

    # 事件注册
    def register(self, type: str, handler: HandlerType) -> None:
        if type == EVENT_BAR:
            handler_list: list = self._bar_handlers[type]
        elif type == EVENT_TICK:
            handler_list: list = self._tick_handlers[type]
        elif type == EVENT_ORDER:
            handler_list: list = self._order_handlers[type]
        else:
            handler_list: list = self._time_handlers[type]
        if handler not in handler_list:
            handler_list.append(handler)

    # 事件注销
    def unregister(self, type: str, handler: HandlerType) -> None:
        if type == EVENT_BAR:
            handler_list: list = self._bar_handlers[type]
        elif type == EVENT_TICK:
            handler_list: list = self._tick_handlers[type]
        elif type == EVENT_ORDER:
            handler_list: list = self._order_handlers[type]
        else:
            handler_list: list = self._time_handlers[type]

        if handler in handler_list:
            handler_list.remove(handler)



"""
1. 只支持单币种保证金模式
2. 只支持全仓模式
3. 只支持单向持仓模式
"""
import base64
import hashlib
import hmac
import json
import time
from copy import copy
from datetime import datetime, timedelta
from threading import Lock
from types import TracebackType
from typing import Any, Dict, List

# 实盘 API地址
from urllib.parse import urlencode

from strategy_api.event.engine import EventEngine
from strategy_api.tm_api.api.rest.rest_client import RestClient, Request, Response
from strategy_api.tm_api.api.websocket.websocket_client import WebsocketClient
from strategy_api.tm_api.base import BaseGateway
from strategy_api.tm_api.object import Status, OrderType, Direction, Interval, TickData, BarData, Product, ContractData, \
    Exchange, DataType, HistoryRequest, PositionSide, OrderRequest, Offset, OrderData, CancelRequest
from strategy_api.tm_api.tools import get_order_type, symbol_deal

REST_HOST: str = "https://www.okx.com"

# 实盘Websocket API地址
PUBLIC_WEBSOCKET_HOST: str = "wss://ws.okx.com:8443/ws/v5/public"
PRIVATE_WEBSOCKET_HOST: str = "wss://ws.okx.com:8443/ws/v5/private"

# 委托状态映射
STATUS_OKEX2VT: Dict[str, Status] = {
    "live": Status.NOTTRADED,
    "effective": Status.ALLTRADED,
    "partially_filled": Status.PARTTRADED,
    "filled": Status.ALLTRADED,
    "canceled": Status.CANCELLED
}

# 委托类型映射
ORDERTYPE_OKEX2VT: Dict[str, OrderType] = {
    "limit": OrderType.LIMIT,
    "market": OrderType.MARKET,
    "conditional": OrderType.MARKET,
    "oco": OrderType.STOP_LOSS_PROFIT
}
ORDERTYPE_VT2OKEX: Dict[OrderType, str] = {v: k for k, v in ORDERTYPE_OKEX2VT.items()}

# 买卖方向映射
DIRECTION_OKEX2VT: Dict[str, Direction] = {
    "buy": Direction.LONG,
    "sell": Direction.SHORT
}
DIRECTION_VT2OKEX: Dict[Direction, str] = {v: k for k, v in DIRECTION_OKEX2VT.items()}

# 合约数据全局缓存字典
symbol_contract_map: Dict[str, ContractData] = {}

INTERVAL_VT2OKEX: Dict[Interval, str] = {
    Interval.MINUTE: "1m",
    Interval.MINUTE_3: "3m",
    Interval.MINUTE_5: "5m",
    Interval.MINUTE_15: "15m",
    Interval.MINUTE_30: "30m",
    Interval.HOUR: "1H",
    Interval.HOUR_2: "2H",
    Interval.HOUR_4: "4H",
}


class OkexFutureUsdtGateway(BaseGateway):
    """
    vn.py用于对接OKEX统一账户的交易接口。
    """

    default_setting: Dict[str, Any] = {
        "key": "",
        "secret": "",
        "Passphrase": "",
        "proxy_host": "",
        "proxy_port": 0,
    }
    exchange: Exchange = Exchange.OKEX
    product: Product = Product.U_FUTURES
    symbolMap: dict = dict()

    def __init__(self, event_engine: EventEngine, order_switch: bool = False, bar_switch: bool = False,
                 tick_switch: bool = False) -> None:
        """init"""
        super().__init__(event_engine, order_switch, bar_switch, tick_switch)

        self.market_ws_api: "OkexWebsocketPublicApi" = OkexWebsocketPublicApi(self)
        self.trade_ws_api: "OkexWebsocketPrivateApi" = OkexWebsocketPrivateApi(self)
        self.rest_api: "OkexRestApi" = OkexRestApi(self)

        self.fail_send_order: int = 0

    def connect(self, setting: dict) -> None:
        """连接交易接口"""
        key: str = setting["key"]
        secret: str = setting["secret"]
        passphrase: str = setting["Passphrase"]

        if isinstance(setting["proxy_host"], str):
            proxy_host: str = setting["proxy_host"]
        else:
            proxy_host: str = ""

        if isinstance(setting["proxy_port"], int):
            proxy_port: int = setting["proxy_port"]
        else:
            proxy_port: int = 0

        self.rest_api.connect(
            key,
            secret,
            passphrase,
            proxy_host,
            proxy_port,
        )

        self.market_ws_api.connect(
            proxy_host,
            proxy_port
        )

        self.trade_ws_api.connect(
            key,
            secret,
            passphrase,
            proxy_host,
            proxy_port,
        )

    # 订阅数据
    def subscribe(self, symbol: str, data_type: DataType, interval: Interval = None) -> None:
        new_symbol = symbol_deal(symbol, self.exchange)
        self.symbolMap[new_symbol] = symbol

        if data_type == DataType.TICK:
            self.market_ws_api.subscribe(new_symbol)
        elif data_type == DataType.BAR:
            if interval:
                self.market_ws_api.subscribe(new_symbol, interval)
            else:
                print(f"订阅失败，k 线数据未定义间隔时间")
        else:
            print(f"订阅失败，未知数据类型: {data_type}")

    # 获取历史K线
    def query_history(self, symbol: str, interval: Interval, hour: int = 0, minutes: int = 0) -> List[BarData]:
        new_symbol = symbol_deal(symbol, self.exchange)
        self.symbolMap[new_symbol] = symbol

        end: datetime = datetime.now()
        if hour:
            start: datetime = end - timedelta(hours=hour + 1)
        elif minutes:
            start: datetime = end - timedelta(minutes=minutes + 1)
        else:
            print("分钟，小时参数都为0, 查询k线失败")
            return []

        req: HistoryRequest = HistoryRequest(
            symbol=new_symbol,
            start=start,
            end=end,
            interval=interval
        )
        return self.rest_api.query_history(req)

    def new_order_id(self) -> str:
        return self.rest_api.get_order_id()

    # 多头买进
    def buy(self,
            orderid: str,
            symbol: str,
            volume: float,  # 数量
            price: float = 0,  # 价格
            maker: bool = False,  # 限价单
            stop_loss: bool = False,  # 止损
            stop_loss_price: float = 0,  # 止损价
            stop_profit: bool = False,  # 止盈
            stop_profit_price: float = 0,  # 止盈价
            position_side: PositionSide = PositionSide.ONEWAY,
            ):

        new_symbol = symbol_deal(symbol, self.exchange)
        self.symbolMap[new_symbol] = symbol

        original_req: OrderRequest = OrderRequest(
            symbol=new_symbol,
            direction=Direction.LONG,
            offset=Offset.OPEN,
            type=get_order_type(maker, stop_loss, stop_profit, self.exchange),
            price=price,
            volume=volume,
            positionSide=position_side,
            exchange=self.exchange,
            stop_loss_price=stop_loss_price,
            stop_profit_price=stop_profit_price
        )
        return self.send_order(original_req, orderid)

    # 多头卖出
    def sell(self,
             orderid: str,
             symbol: str,
             volume: float,  # 数量
             price: float = 0,  # 价格
             maker: bool = False,  # 限价单
             stop_loss: bool = False,  # 止损
             stop_loss_price: float = 0,  # 止损价
             stop_profit: bool = False,  # 止盈
             stop_profit_price: float = 0,  # 止盈价
             position_side: PositionSide = PositionSide.ONEWAY,
             ):
        new_symbol = symbol_deal(symbol, self.exchange)
        self.symbolMap[new_symbol] = symbol

        original_req: OrderRequest = OrderRequest(
            symbol=new_symbol,
            direction=Direction.SHORT,
            offset=Offset.CLOSE,
            type=get_order_type(maker, stop_loss, stop_profit, self.exchange),
            price=price,
            volume=volume,
            positionSide=position_side,
            exchange=self.exchange,
            stop_loss_price=stop_loss_price,
            stop_profit_price=stop_profit_price
        )
        return self.send_order(original_req, orderid)

    # 空头买进
    def short(self,
              orderid: str,
              symbol: str,
              volume: float,  # 数量
              price: float = 0,  # 价格
              maker: bool = False,  # 限价单
              stop_loss: bool = False,  # 止损
              stop_loss_price: float = 0,  # 止损价
              stop_profit: bool = False,  # 止盈
              stop_profit_price: float = 0,  # 止盈价
              position_side: PositionSide = PositionSide.ONEWAY,
              ):
        new_symbol = symbol_deal(symbol, self.exchange)
        self.symbolMap[new_symbol] = symbol

        original_req: OrderRequest = OrderRequest(
            symbol=new_symbol,
            direction=Direction.SHORT,
            offset=Offset.OPEN,
            type=get_order_type(maker, stop_loss, stop_profit, self.exchange),
            price=price,
            volume=volume,
            positionSide=position_side,
            exchange=self.exchange,
            stop_loss_price=stop_loss_price,
            stop_profit_price=stop_profit_price
        )
        return self.send_order(original_req, orderid)

    # 空头卖出
    def cover(self,
              orderid: str,
              symbol: str,
              volume: float,  # 数量
              price: float = 0,  # 价格
              maker: bool = False,  # 限价单
              stop_loss: bool = False,  # 止损
              stop_loss_price: float = 0,  # 止损价
              stop_profit: bool = False,  # 止盈
              stop_profit_price: float = 0,  # 止盈价
              position_side: PositionSide = PositionSide.ONEWAY,
              ):

        new_symbol = symbol_deal(symbol, self.exchange)
        self.symbolMap[new_symbol] = symbol

        original_req: OrderRequest = OrderRequest(
            symbol=new_symbol,
            direction=Direction.LONG,
            offset=Offset.CLOSE,
            type=get_order_type(maker, stop_loss, stop_profit, self.exchange),
            price=price,
            volume=volume,
            positionSide=position_side,
            exchange=self.exchange,
            stop_loss_price=stop_loss_price,
            stop_profit_price=stop_profit_price
        )
        return self.send_order(original_req, orderid)

    # 发送订单
    def send_order(self, req: OrderRequest, orderid: str):
        self.fail_send_order = 0
        self.rest_api.send_order(req, orderid)

    # 撤销订单
    def cancel_order(self, orderid: str, symbol: str) -> None:
        new_symbol = symbol_deal(symbol, self.exchange)
        self.symbolMap[new_symbol] = symbol

        req: CancelRequest = CancelRequest(
            orderid=orderid,
            symbol=new_symbol,
            exchange=self.exchange
        )
        self.rest_api.cancel_order(req)

    def close(self) -> None:
        """关闭连接"""
        self.rest_api.stop()
        self.trade_ws_api.stop()
        self.market_ws_api.stop()


class OkexRestApi(RestClient):
    """"""

    def __init__(self, gateway: OkexFutureUsdtGateway) -> None:
        """构造函数"""
        super().__init__()

        self.gateway: OkexFutureUsdtGateway = gateway

        self.key: str = ""
        self.secret: str = ""
        self.passphrase: str = ""
        self.simulated: bool = False

        self.order_count: int = 0
        self.connect_time: int = 0

    def sign(self, request: Request) -> Request:
        """生成欧易V5签名"""
        # 签名
        timestamp: str = generate_timestamp()
        request.data = json.dumps(request.data)

        if request.params:
            path: str = request.path + "?" + urlencode(request.params)
        else:
            path: str = request.path

        msg: str = timestamp + request.method + path + request.data
        signature: bytes = generate_signature(msg, self.secret)

        # 添加请求头
        request.headers = {
            "OK-ACCESS-KEY": self.key,
            "OK-ACCESS-SIGN": signature.decode(),
            "OK-ACCESS-TIMESTAMP": timestamp,
            "OK-ACCESS-PASSPHRASE": self.passphrase,
            "Content-Type": "application/json"
        }

        if self.simulated:
            request.headers["x-simulated-trading"] = "1"

        return request

    def connect(
            self,
            key: str,
            secret: str,
            passphrase: str,
            proxy_host: str,
            proxy_port: int,
    ) -> None:
        """连接REST服务器"""
        self.key = key
        self.secret = secret.encode()
        self.passphrase = passphrase

        self.connect_time = int(datetime.now().strftime("%y%m%d%H%M%S"))

        self.init(REST_HOST, proxy_host, proxy_port)
        self.start()
        print("OKEX REST API启动成功")

        self.query_time()

    def query_time(self) -> None:
        """查询时间"""
        self.add_request(
            "GET",
            "/api/v5/public/time",
            callback=self.on_query_time
        )

    def on_query_time(self, packet: dict, request: Request) -> None:
        """时间查询回报"""
        timestamp: int = int(packet["data"][0]["ts"])
        server_time: datetime = datetime.fromtimestamp(timestamp / 1000)
        local_time: datetime = datetime.now()
        msg: str = f"服务器时间：{server_time}，本机时间：{local_time}"

    def on_error(
            self,
            exception_type: type,
            exception_value: Exception,
            tb: TracebackType,
            request: Request
    ) -> None:
        """触发异常回报"""
        msg: str = f"触发异常，状态码：{exception_type}，信息：{exception_value}"
        print(msg)

    def query_history(self, req: HistoryRequest) -> List[BarData]:
        """
        查询历史数据
        K线数据每个粒度最多可获取最近1440条
        """
        buf: Dict[datetime, BarData] = {}
        start_time: str = str(int(time.mktime(req.start.timetuple()) * 1000))
        end_time: str = str(int(time.mktime(req.end.timetuple()) * 1000))

        limit: int = 300
        path: str = "/api/v5/market/candles"

        while True:
            # 创建查询参数
            params: dict = {
                "instId": req.symbol,
                "bar": INTERVAL_VT2OKEX[req.interval],
                "limit": limit
            }

            if end_time:
                params["after"] = end_time
            if start_time:
                params["before"] = start_time

            # 从服务器获取响应
            resp: Response = self.request(
                "GET",
                path,
                params=params
            )

            # 如果请求失败则终止循环
            if resp.status_code // 100 != 2:
                msg = f"获取历史数据失败，状态码：{resp.status_code}，信息：{resp.text}"
                print(msg)
                break
            else:
                data: dict = resp.json()

                if not data["data"]:
                    m = data["msg"]
                    msg = f"获取历史数据为空，{m}"
                    print(msg)
                    break

                for bar_list in data["data"]:
                    bar: BarData = BarData(
                        symbol=self.gateway.symbolMap[req.symbol],
                        exchange=self.gateway.exchange,
                        volume=float(bar_list[5]),
                        turnover=float(bar_list[7]),
                        open_price=float(bar_list[1]),
                        high_price=float(bar_list[2]),
                        low_price=float(bar_list[3]),
                        close_price=float(bar_list[4]),
                        datetime=generate_datetime(bar_list[0]),
                        endTime=generate_datetime(bar_list[0])
                    )
                    bar.endTime = None
                    buf[bar.datetime] = bar

                begin: str = data["data"][-1][0]
                end: str = data["data"][0][0]
                msg: str = f"获取历史数据成功，{req.symbol} - {req.interval.value}，{generate_datetime(begin)} - {generate_datetime(end)}"
                if len(data) < limit:
                    break
                # # 更新结束时间
                end_time = begin
            time.sleep(0.5)

        index: List[datetime] = list(buf.keys())
        index.sort()

        history: List[BarData] = [buf[i] for i in index]
        return history

    def get_order_id(self) -> str:
        self.order_count += 1
        count_str = str(self.order_count).rjust(6, "0")
        orderid = f"{self.connect_time}{count_str}"
        return orderid

    def send_order(self, req: OrderRequest, orderid: str):

        order: OrderData = req.create_order_data(
            orderid,
        )
        self.gateway.on_order(order)
        if req.positionSide == PositionSide.ONEWAY:
            params: dict = {
                "instId": req.symbol,  # 标的
                "tdMode": "cross",  # 交易模式
                "clOrdId": orderid,  # 用户定义的ID
                "side": DIRECTION_VT2OKEX[req.direction],  # 方向
                "sz": str(req.volume)  # 数量
            }
            if req.offset == Offset.CLOSE:
                params['reduceOnly'] = True
        else:
            if (req.direction == Direction.SHORT and req.offset == Offset.OPEN) or (
                    req.direction == Direction.LONG and req.offset == Offset.CLOSE):
                position_side = "short"
            else:
                position_side = "long"
            params: dict = {
                "instId": req.symbol,  # 标的
                "posSide": position_side,
                "tdMode": "cross",  # 交易模式
                "clOrdId": orderid,  # 用户定义的ID
                "side": DIRECTION_VT2OKEX[req.direction],  # 方向
                "sz": str(req.volume)  # 数量
            }
        if req.type == OrderType.LIMIT:
            params['ordType'] = "limit"
            params['px'] = req.price  # 价格
        else:
            params['ordType'] = "market"

        path: str = "/api/v5/trade/order"

        if req.type == OrderType.STOP_LOSS_PROFIT:
            params['slTriggerPx'] = str(req.stop_loss_price)  # 止损触发价
            params['slOrdPx'] = '-1'  # 止损委托价,当为 -1 时。按市价止损
            params['tpTriggerPx'] = str(req.stop_profit_price)  # 止盈触发价
            params['tpOrdPx'] = '-1'  # 止盈委托价, 当为-1 时，按市价止盈
            params['ordType'] = "oco"
            path: str = "/api/v5/trade/order-algo"

        elif req.type == OrderType.STOP_MARKET:
            params['slTriggerPx'] = str(req.price)  # 止损触发价
            params['slOrdPx'] = '-1'  # 止损委托价,当为 -1 时。按市价止损
            params['ordType'] = "conditional"
            path: str = "/api/v5/trade/order-algo"

        elif req.type == OrderType.TAKE_PROFIT_MARKET:
            params['tpTriggerPx'] = str(req.price)  # 止盈触发价
            params['tpOrdPx'] = '-1'  # 止盈委托价, 当为-1 时，按市价止盈
            params['ordType'] = "conditional"
            path: str = "/api/v5/trade/order-algo"

        self.add_request(
            method="POST",
            path=path,
            callback=self.on_send_order,
            data=params,
            extra={'order': order, 'req': req, 'orderid': orderid},
            on_error=self.on_send_order_error,
            on_failed=self.on_send_order_failed
        )

    def cancel_order(self, req: CancelRequest) -> None:
        data: dict = {
            "instId": req.symbol,
            "clOrdId": req.orderid,
        }

        path: str = "/api/v5/trade/cancel-order"

        self.add_request(
            method="POST",
            path=path,
            callback=self.on_cancel_order,
            data=data,
            on_failed=self.on_cancel_order_failed,
        )

    # 发送订单成功回调
    def on_send_order(self, data: dict, request: Request) -> None:
        if data['code'] != "0":
            print(f"订单被拒绝: {data}")
            if request.extra:
                order: OrderData = copy(request.extra['order'])
                order.status = Status.REJECTED
                self.gateway.on_order(order)
                print(request)

    # 发送订单失败回调
    def on_send_order_failed(self, status_code: int, request: Request) -> None:
        print(f"发送订单 失败：{status_code}, {request}")
        pass

    # 发送订单错误回调
    def on_send_order_error(
            self, exception_type: type, exception_value: Exception, tb, request: Request
    ) -> None:
        if self.gateway.fail_send_order < 3:
            req: OrderRequest = copy(request.extra['req'])
            orderid: str = copy(request.extra['orderid'])
            self.send_order(req, orderid)
            self.gateway.fail_send_order += 1
            print(f"发送订单错误,重新发送, {request}")
        print(f"发送订单错误, {request}")

    # 撤销订单成功回调
    def on_cancel_order(self, data: dict, request: Request) -> None:
        print(f"撤销订单成功: {request}")

    # 撤销订单失败
    def on_cancel_order_failed(self, status_code: int, request: Request) -> None:
        print(f"撤销订单失败: {request}")


class OkexWebsocketPublicApi(WebsocketClient):
    """"""

    def __init__(self, gateway: OkexFutureUsdtGateway) -> None:
        """构造函数"""
        super().__init__()

        self.gateway: OkexFutureUsdtGateway = gateway
        self.ticks: Dict[str, TickData] = {}
        self.bars: Dict[str, BarData] = {}

    def connect(
            self,
            proxy_host: str,
            proxy_port: int,
    ) -> None:
        """连接Websocket公共频道"""
        self.init(PUBLIC_WEBSOCKET_HOST, proxy_host, proxy_port, 20)
        self.start()

    def subscribe(self, symbol: str, interval: Interval = None) -> None:
        """订阅行情"""
        if interval:
            target = symbol + f"*{interval.value}"
            if target in self.bars:
                return
            bar: BarData = BarData(
                symbol=symbol,
                datetime=datetime.now(),
                endTime=datetime.now(),
                interval=interval,
                exchange=self.gateway.exchange,
                product=self.gateway.product
            )
            self.bars[target] = bar
            args = [
                {"channel": f"candle{interval.value}", "instId": symbol}
            ]
        else:
            if symbol in self.ticks:
                return
            tick: TickData = TickData(
                symbol=symbol,
                datetime=datetime.now(),
                exchange=self.gateway.exchange,
                product=self.gateway.product
            )
            self.ticks[symbol] = tick
            args = [
                {"channel": "tickers", "instId": symbol}
            ]

        req: dict = {
            "op": "subscribe",
            "args": args
        }
        self.send_packet(req)

    def on_connected(self) -> None:
        """连接成功回报"""
        args = []
        for target in self.bars:
            symbol, interval = target.split("*")
            d = {"channel": f"candle{interval}", "instId": symbol}
            args.append(d)
        for target in self.ticks:
            d = {"channel": "tickers", "instId": target}
            args.append(d)
        if args:
            req: dict = {
                "op": "subscribe",
                "args": args
            }
            self.send_packet(req)
        print("Okex Websocket Public API连接成功")
        pass

    def on_disconnected(self) -> None:
        """连接断开回报"""
        pass
        # print("Okex Websocket Public API连接断开")

    def on_packet(self, packet: dict) -> None:
        """推送数据回报"""
        if "event" in packet:
            event: str = packet["event"]
            if event == "subscribe":
                return
            elif event == "error":
                code: str = packet["code"]
                msg: str = packet["msg"]
                print(f"Okex Websocket Public API请求异常, 状态码：{code}, 信息：{msg}")
        else:
            channel: str = packet["arg"]["channel"]
            symbol: str = packet["arg"].get("instId")
            data: list = packet['data']
            if channel == "tickers":
                for d in data:
                    tick: TickData = self.ticks[symbol]
                    tick.last_price = float(d["last"])
                    tick.open_price = float(d["open24h"])
                    tick.high_price = float(d["high24h"])
                    tick.low_price = float(d["low24h"])
                    tick.volume = float(d["vol24h"])
                    if tick.last_price:
                        self.gateway.on_tick(copy(tick))

            elif channel.startswith("candle"):
                for d in data:
                    if d[8] == "1":
                        bar: BarData = self.bars[symbol + f"*{channel.split('e')[-1]}"]
                        bar.volume = float(d[5])
                        bar.turnover = float(d[7])
                        bar.open_price = float(d[1])
                        bar.high_price = float(d[2])
                        bar.low_price = float(d[3])
                        bar.close_price = float(d[4])
                        bar.datetime = generate_datetime(str(d[0]))
                        bar.endTime = None
                        self.gateway.on_bar(copy(bar))

    def on_error(self, exception_type: type, exception_value: Exception, tb) -> None:
        """触发异常回报"""
        msg: str = f"Okex 公共频道触发异常，类型：{exception_type}，信息：{exception_value}"
        print(msg)


class OkexWebsocketPrivateApi(WebsocketClient):
    """"""

    def __init__(self, gateway: OkexFutureUsdtGateway) -> None:
        """构造函数"""
        super().__init__()

        self.gateway: OkexFutureUsdtGateway = gateway

        self.key: str = ""
        self.secret: str = ""
        self.passphrase: str = ""

        self.reqid: int = 0
        self.order_count: int = 0
        self.connect_time: int = 0

        self.callbacks: Dict[str, callable] = {
            "login": self.on_login,
            "orders": self.on_order,
            "orders-algo": self.on_order,
            "error": self.on_api_error
        }

        self.reqid_order_map: Dict[str, OrderData] = {}

    def connect(
            self,
            key: str,
            secret: str,
            passphrase: str,
            proxy_host: str,
            proxy_port: int
    ) -> None:
        """连接Websocket私有频道"""
        self.key = key
        self.secret = secret.encode()
        self.passphrase = passphrase
        self.connect_time = int(datetime.now().strftime("%y%m%d%H%M%S"))
        self.init(PRIVATE_WEBSOCKET_HOST, proxy_host, proxy_port, 20)
        self.start()

    def on_connected(self) -> None:
        """连接成功回报"""
        self.login()

    def on_disconnected(self) -> None:
        """连接断开回报"""
        pass
        # print("Websocket Private API连接断开")

    def on_packet(self, packet: dict) -> None:
        """推送数据回报"""
        if "event" in packet:
            cb_name: str = packet["event"]
        elif "op" in packet:
            cb_name: str = packet["op"]
        else:
            cb_name: str = packet["arg"]["channel"]

        callback: callable = self.callbacks.get(cb_name, None)
        if callback:
            callback(packet)

    def on_error(self, exception_type: type, exception_value: Exception, tb) -> None:
        """触发异常回报"""
        msg: str = f"私有频道触发异常，类型：{exception_type}，信息：{exception_value}"
        print(msg)

    def on_api_error(self, packet: dict) -> None:
        """用户登录请求回报"""
        code: str = packet["code"]
        msg: str = packet["msg"]
        print(f"Websocket Private API请求失败, 状态码：{code}, 信息：{msg}")

    def on_login(self, packet: dict) -> None:
        """用户登录请求回报"""
        if packet["code"] == '0':
            self.subscribe_topic()
        else:
            print("Websocket Private API登录失败")

    def login(self) -> None:
        """用户登录"""
        timestamp: str = str(time.time())
        msg: str = timestamp + "GET" + "/users/self/verify"
        signature: bytes = generate_signature(msg, self.secret)

        okex_req: dict = {
            "op": "login",
            "args":
                [
                    {
                        "apiKey": self.key,
                        "passphrase": self.passphrase,
                        "timestamp": timestamp,
                        "sign": signature.decode("utf-8")
                    }
                ]
        }
        self.send_packet(okex_req)

    def on_order(self, packet: dict) -> None:
        """委托更新推送"""
        data: list = packet["data"]
        for d in data:
            order: OrderData = parse_order_data(d)
            self.gateway.on_order(order)

    def subscribe_topic(self) -> None:
        """订阅委托推送"""
        okex_req: dict = {
            "op": "subscribe",
            "args": [
                {
                    "channel": "orders",
                    "instType": "ANY"
                },
                {
                    "channel": "orders-algo",
                    "instType": "ANY"
                },
            ]
        }
        self.send_packet(okex_req)


def generate_datetime(timestamp: str) -> datetime:
    """generate time"""
    dt: datetime = datetime.fromtimestamp(int(timestamp) / 1000)
    return dt


def generate_timestamp() -> str:
    """生成时间戳"""
    now: datetime = datetime.utcnow()
    timestamp: str = now.isoformat("T", "milliseconds")
    return timestamp + "Z"


def generate_signature(msg: str, secret_key: str) -> bytes:
    """生成签名"""
    return base64.b64encode(hmac.new(secret_key, msg.encode(), hashlib.sha256).digest())


def parse_order_data(data: dict) -> OrderData:
    """解析委托回报数据"""
    order_id: str = data["clOrdId"]

    order: OrderData = OrderData(
        symbol=data["instId"],
        exchange=Exchange.OKEX,
        type=ORDERTYPE_OKEX2VT[data["ordType"]],
        orderid=order_id,
        direction=DIRECTION_OKEX2VT[data["side"]],
        offset=Offset.NONE,
        traded=0,
        price=0,
        volume=0,
        datetime=generate_datetime(data["cTime"]),
        status=STATUS_OKEX2VT[data["state"]],
    )
    # if data['px'] != "":
    #     order.price = float(data['px'])
    # else:
    #     order.price = None
    return order



import sys
import traceback
from datetime import datetime
from typing import Any, Callable, Optional, Union, Type
from types import TracebackType, coroutine
from threading import Thread
from asyncio import (
    get_event_loop,
    set_event_loop,
    run_coroutine_threadsafe,
    AbstractEventLoop,
    Future
)
from json import loads

from aiohttp import ClientSession, ClientResponse

CALLBACK_TYPE = Callable[[Union[dict, list], "Request"], None]
ON_FAILED_TYPE = Callable[[int, "Request"], None]
ON_ERROR_TYPE = Callable[[Type, Exception, TracebackType, "Request"], None]


class Request(object):
    def __init__(
            self,
            method: str,
            path: str,
            params: dict,
            data: Union[dict, str, bytes],
            headers: dict,
            callback: CALLBACK_TYPE = None,
            on_failed: ON_FAILED_TYPE = None,
            on_error: ON_ERROR_TYPE = None,
            extra: Any = None,
    ):
        """"""
        self.method: str = method
        self.path: str = path
        self.callback: CALLBACK_TYPE = callback
        self.params: dict = params
        self.data: Union[dict, str, bytes] = data
        self.headers: dict = headers

        self.on_failed: ON_FAILED_TYPE = on_failed
        self.on_error: ON_ERROR_TYPE = on_error
        self.extra: Any = extra

        self.response: "Response" = None

    def __str__(self):
        if self.response is None:
            status_code = "terminated"
        else:
            status_code = self.response.status_code

        return (
            "request : {} {} because {}: \n"
            "headers: {}\n"
            "params: {}\n"
            "data: {}\n"
            "response:"
            "{}\n".format(
                self.method,
                self.path,
                status_code,
                self.headers,
                self.params,
                self.data,
                "" if self.response is None else self.response.text,
            )
        )


class Response:

    def __init__(self, status_code: int, text: str) -> None:
        """"""
        self.status_code: int = status_code
        self.text: str = text

    def json(self) -> dict:
        """convert str into json format"""
        data = loads(self.text)
        return data


class RestClient(object):

    def __init__(self):
        """"""
        self.url_base: str = ""
        self.proxy: str = ""

        self.session: ClientSession = ClientSession(trust_env=True)
        self.loop: AbstractEventLoop = None

    def init(
            self,
            url_base: str,
            proxy_host: str = "",
            proxy_port: int = 0
    ) -> None:
        """base url or api end point"""
        self.url_base = url_base

        if proxy_host and proxy_port:
            self.proxy = f"http://{proxy_host}:{proxy_port}"

    def start(self) -> None:
        """start event loop"""
        if not self.loop:
            self.loop = get_event_loop()

        start_event_loop(self.loop)

    def stop(self) -> None:
        if self.loop and self.loop.is_running():
            self.loop.stop()
            print("rest 停止")

    def join(self) -> None:
        pass

    def add_request(
            self,
            method: str,
            path: str,
            callback: CALLBACK_TYPE,
            params: dict = None,
            data: Union[dict, str, bytes] = None,
            headers: dict = None,
            on_failed: ON_FAILED_TYPE = None,
            on_error: ON_ERROR_TYPE = None,
            extra: Any = None,
    ) -> Request:
        request: Request = Request(
            method,
            path,
            params,
            data,
            headers,
            callback,
            on_failed,
            on_error,
            extra,
        )

        coro: coroutine = self._process_request(request)
        if self.loop:
            run_coroutine_threadsafe(coro, self.loop)
        return request

    def request(
            self,
            method: str,
            path: str,
            params: dict = None,
            data: dict = None,
            headers: dict = None,
    ) -> Response:
        request: Request = Request(method, path, params, data, headers)
        coro: coroutine = self._get_response(request)
        fut: Future = run_coroutine_threadsafe(coro, self.loop)
        return fut.result()

    def sign(self, request: Request) -> Request:
        return request

    def on_failed(self, status_code: int, request: Request) -> None:
        print(f"RestClient on failed " + "-" * 10)
        print(str(request))

    def on_error(
            self,
            exception_type: type,
            exception_value: Exception,
            tb,
            request: Optional[Request],
    ) -> None:
        try:
            print("RestClient 异常 " + "-" * 10)
            print(self.exception_detail(exception_type, exception_value, tb, request))
        except Exception:
            traceback.print_exc()

    def exception_detail(
            self,
            exception_type: type,
            exception_value: Exception,
            tb,
            request: Optional[Request],
    ) -> str:
        text = "[{}]: Unhandled RestClient Error:{}\n".format(
            datetime.now().isoformat(), exception_type
        )
        text += "request:{}\n".format(request)
        text += "Exception trace: \n"
        text += "".join(
            traceback.format_exception(exception_type, exception_value, tb)
        )
        return text

    async def _get_response(self, request: Request) -> Response:
        request = self.sign(request)
        url = self._make_full_url(request.path)
        if self.proxy:
            cr: ClientResponse = await self.session.request(
                request.method,
                url,
                headers=request.headers,
                params=request.params,
                data=request.data,
                proxy=self.proxy,
                verify_ssl=False
            )
        else:
            cr: ClientResponse = await self.session.request(
                request.method,
                url,
                headers=request.headers,
                params=request.params,
                data=request.data
            )

        text: str = await cr.text()
        status_code = cr.status

        request.response = Response(status_code, text)
        return request.response

    async def _process_request(self, request: Request) -> None:
        try:
            response: Response = await self._get_response(request)
            status_code: int = response.status_code

            # 2xx code means success
            if status_code // 100 == 2:
                request.callback(response.json(), request)
            # failed
            else:
                # if on_failed request, then call the callback
                if request.on_failed:
                    request.on_failed(status_code, request)

                else:
                    self.on_failed(status_code, request)
        except Exception:
            t, v, tb = sys.exc_info()
            # on error.
            if request.on_error:
                request.on_error(t, v, tb, request)

            else:
                self.on_error(t, v, tb, request)

    def _make_full_url(self, path: str) -> str:
        url: str = self.url_base + path
        return url


def start_event_loop(loop: AbstractEventLoop) -> None:
    if not loop.is_running():
        thread = Thread(target=run_event_loop, args=(loop,))
        thread.daemon = True
        thread.start()


def run_event_loop(loop: AbstractEventLoop) -> None:
    set_event_loop(loop)
    loop.run_forever()



# rest api host
import hashlib
import hmac
import json
import time
import urllib
from copy import copy
from enum import Enum
from ssl import SSLError
from threading import Lock
from datetime import datetime, timedelta
from typing import Dict, Any, Tuple, List
from strategy_api.event.engine import EVENT_TIMER, Event, EventEngine
from strategy_api.tm_api.api.rest.rest_client import RestClient, Request, Response
from strategy_api.tm_api.api.websocket.websocket_client import WebsocketClient
from strategy_api.tm_api.base import BaseGateway
from strategy_api.tm_api.object import TickData, BarData, Interval, OrderRequest, OrderData, Direction, OrderType, \
    Offset, Status, CancelRequest, HistoryRequest, PositionSide, DataType, Exchange, Product

# http 基础地址
from strategy_api.tm_api.tools import get_order_type, symbol_deal

F_REST_HOST: str = "https://dapi.binance.com"

# ws 基础地址
F_WEBSOCKET_TRADE_HOST: str = "wss://dstream.binance.com/ws/"
F_WEBSOCKET_DATA_HOST: str = "wss://dstream.binance.com/stream"

# sell/buy direction map
DIRECTION_VT2BINANCES: Dict[Direction, str] = {
    Direction.LONG: "BUY",
    Direction.SHORT: "SELL"
}

DIRECTION_BINANCES2VT: Dict[str, Direction] = {v: k for k, v in DIRECTION_VT2BINANCES.items()}

# GTC 成交为止, 一直有效
ORDERTYPE_VT2BINANCES: Dict[OrderType, Tuple[str, str]] = {
    OrderType.LIMIT: ("LIMIT", "GTC"),  # 限价单
    OrderType.MARKET: ("MARKET", "GTC"),  # 市价单
    OrderType.STOP_MARKET: ("STOP_MARKET", "GTC"),  # 止损单
    OrderType.TAKE_PROFIT_MARKET: ("TAKE_PROFIT_MARKET", "GTC")  # 止盈单
}
ORDERTYPE_BINANCES2VT: Dict[Tuple[str, str], OrderType] = {v: k for k, v in ORDERTYPE_VT2BINANCES.items()}

# Order status map
STATUS_BINANCES2VT: Dict[str, Status] = {
    "NEW": Status.NOTTRADED,
    "PARTIALLY_FILLED": Status.PARTTRADED,
    "FILLED": Status.ALLTRADED,
    "CANCELED": Status.CANCELLED,
    "REJECTED": Status.REJECTED,
    "EXPIRED": Status.CANCELLED
}

TIMEDELTA_MAP: Dict[Interval, timedelta] = {
    Interval.MINUTE: timedelta(minutes=1),
    Interval.HOUR: timedelta(hours=1)
}

INTERVAL_VT2BINANCES: Dict[Interval, str] = {
    Interval.MINUTE: "1m",
    Interval.MINUTE_3: "3m",
    Interval.MINUTE_5: "5m",
    Interval.MINUTE_15: "15m",
    Interval.MINUTE_30: "30m",
    Interval.HOUR: "1h",
    Interval.HOUR_2: "2h",
    Interval.HOUR_4: "4h",
    Interval.HOUR_6: "6h",
    Interval.HOUR_8: "8h",
    Interval.HOUR_12: "12h"
}


# 私钥使用枚举
class Security(Enum):
    NONE: int = 0
    SIGNED: int = 1
    API_KEY: int = 2


class BinanceFutureInverseGateway(BaseGateway):
    default_setting: Dict[str, Any] = {
        "key": "",
        "secret": "",
        "proxy_host": "",
        "proxy_port": 0,
    }
    exchange: Exchange = Exchange.BINANCE
    product: Product = Product.B_FUTURES
    symbolMap: dict = dict()

    def __init__(self, event_engine: EventEngine, order_switch: bool = False, bar_switch: bool = False, tick_switch: bool = False) -> None:
        """init"""
        super().__init__(event_engine, order_switch, bar_switch, tick_switch)
        self.trade_ws_api: "BinanceInverseTradeWebsocketApi" = BinanceInverseTradeWebsocketApi(self)
        self.market_ws_api: "BinanceInverseDataWebsocketApi" = BinanceInverseDataWebsocketApi(self)
        self.rest_api: "BinanceInverseRestApi" = BinanceInverseRestApi(self)
        self.get_server_time_interval: int = 0

    def connect(self, setting: dict) -> None:
        key: str = setting["key"]
        secret: str = setting["secret"]

        if isinstance(setting["proxy_host"], str):
            proxy_host: str = setting["proxy_host"]
        else:
            proxy_host: str = ""

        if isinstance(setting["proxy_port"], int):
            proxy_port: int = setting["proxy_port"]
        else:
            proxy_port: int = 0

        self.rest_api.connect(key, secret, proxy_host, proxy_port)
        self.market_ws_api.connect(proxy_host, proxy_port)

        self.event_engine.unregister(EVENT_TIMER, self.process_timer_event)
        self.event_engine.register(EVENT_TIMER, self.process_timer_event)

    # 定时任务, 不间断获取服务器时间
    def process_timer_event(self, event: Event) -> None:
        self.rest_api.keep_user_stream()
        self.get_server_time_interval += 1
        if self.get_server_time_interval >= 300:
            self.rest_api.query_time()
            self.get_server_time_interval = 0

    def new_order_id(self) -> str:
        return self.rest_api.get_order_id()

    # 多头买进
    def buy(self,
            orderid: str,
            symbol: str,
            volume: float,  # 数量
            price: float = 0,  # 价格
            maker: bool = False,  # 限价单
            stop_loss: bool = False,  # 止损
            stop_loss_price: float = 0,  # 止损价
            stop_profit: bool = False,  # 止盈
            stop_profit_price: float = 0,  # 止盈价
            position_side: PositionSide = PositionSide.ONEWAY,
            ):
        new_symbol = symbol_deal(symbol, self.exchange)
        self.symbolMap[new_symbol] = symbol

        original_req: OrderRequest = OrderRequest(
            symbol=new_symbol,
            direction=Direction.LONG,
            offset=Offset.OPEN,
            type=get_order_type(maker, stop_loss, stop_profit, self.exchange),
            price=price,
            volume=volume,
            positionSide=position_side,
            exchange=self.exchange
        )
        self.send_order(original_req, orderid)

    # 多头卖出
    def sell(self,
             orderid: str,
             symbol: str,
             volume: float,  # 数量
             price: float = 0,  # 价格
             maker: bool = False,  # 限价单
             stop_loss: bool = False,  # 止损
             stop_loss_price: float = 0,  # 止损价
             stop_profit: bool = False,  # 止盈
             stop_profit_price: float = 0,  # 止盈价
             position_side: PositionSide = PositionSide.ONEWAY,
             ):
        new_symbol = symbol_deal(symbol, self.exchange)
        self.symbolMap[new_symbol] = symbol

        original_req: OrderRequest = OrderRequest(
            symbol=new_symbol,
            direction=Direction.SHORT,
            offset=Offset.CLOSE,
            type=get_order_type(maker, stop_loss, stop_profit, self.exchange),
            price=price,
            volume=volume,
            positionSide=position_side,
            exchange=self.exchange
        )
        self.send_order(original_req, orderid)

    # 空头买进
    def short(self,
              orderid: str,
              symbol: str,
              volume: float,  # 数量
              price: float = 0,  # 价格
              maker: bool = False,  # 限价单
              stop_loss: bool = False,  # 止损
              stop_loss_price: float = 0,  # 止损价
              stop_profit: bool = False,  # 止盈
              stop_profit_price: float = 0,  # 止盈价
              position_side: PositionSide = PositionSide.ONEWAY,
              ):
        new_symbol = symbol_deal(symbol, self.exchange)
        self.symbolMap[new_symbol] = symbol

        original_req: OrderRequest = OrderRequest(
            symbol=new_symbol,
            direction=Direction.SHORT,
            offset=Offset.OPEN,
            type=get_order_type(maker, stop_loss, stop_profit, self.exchange),
            price=price,
            volume=volume,
            positionSide=position_side,
            exchange=self.exchange
        )
        self.send_order(original_req, orderid)

    # 空头卖出
    def cover(self,
              orderid: str,
              symbol: str,
              volume: float,  # 数量
              price: float = 0,  # 价格
              maker: bool = False,  # 限价单
              stop_loss: bool = False,  # 止损
              stop_loss_price: float = 0,  # 止损价
              stop_profit: bool = False,  # 止盈
              stop_profit_price: float = 0,  # 止盈价
              position_side: PositionSide = PositionSide.ONEWAY,
              ):
        new_symbol = symbol_deal(symbol, self.exchange)
        self.symbolMap[new_symbol] = symbol

        original_req: OrderRequest = OrderRequest(
            symbol=new_symbol,
            direction=Direction.LONG,
            offset=Offset.CLOSE,
            type=get_order_type(maker, stop_loss, stop_profit, self.exchange),
            price=price,
            volume=volume,
            positionSide=position_side,
            exchange=self.exchange
        )
        self.send_order(original_req, orderid)

    # 发送订单
    def send_order(self, req: OrderRequest, orderid: str):
        return self.rest_api.send_order(req, orderid)

    # 撤销订单
    def cancel_order(self, orderid: str, symbol: str) -> None:
        new_symbol = symbol_deal(symbol, self.exchange)
        self.symbolMap[new_symbol] = symbol

        req: CancelRequest = CancelRequest(
            orderid=orderid,
            symbol=new_symbol,
            exchange=self.exchange
        )
        self.rest_api.cancel_order(req)

    # 获取历史K线
    def query_history(self, symbol: str, interval: Interval, hour: int = 0, minutes: int = 0) -> List[BarData]:
        new_symbol = symbol_deal(symbol, self.exchange)
        self.symbolMap[new_symbol] = symbol

        end: datetime = datetime.now()
        if hour:
            start: datetime = end - timedelta(hours=hour)
        elif minutes:
            start: datetime = end - timedelta(minutes=minutes)
        else:
            print("分钟，小时参数都为0, 查询k线失败")
            return []

        req: HistoryRequest = HistoryRequest(
            symbol=new_symbol,
            start=start,
            end=end,
            interval=interval,
            exchange=self.exchange
        )
        return self.rest_api.query_history(req)

    # 订阅数据
    def subscribe(self, symbol: str, data_type: DataType, interval: Interval = None) -> None:
        new_symbol = symbol_deal(symbol, self.exchange)
        self.symbolMap[new_symbol] = symbol

        if data_type == DataType.TICK:
            self.market_ws_api.subscribe(new_symbol)
        elif data_type == DataType.BAR:
            if interval:
                self.market_ws_api.subscribe(new_symbol, interval)
            else:
                print(f"订阅失败，k 线数据未定义间隔时间")
        else:
            print(f"订阅失败，未知数据类型: {data_type}")

    # 关闭
    def close(self) -> None:
        self.rest_api.stop()
        self.trade_ws_api.stop()
        self.market_ws_api.stop()


class BinanceInverseRestApi(RestClient):

    def __init__(self, gateway: BinanceFutureInverseGateway) -> None:
        super().__init__()

        self.gateway: BinanceFutureInverseGateway = gateway

        self.trade_ws_api: BinanceInverseTradeWebsocketApi = self.gateway.trade_ws_api

        self.key: str = ""
        self.secret: str = ""

        self.user_stream_key: str = ""
        self.keep_alive_count: int = 0
        self.keep_alive_failed_count: int = 0
        self.recv_window: int = 10000
        self.time_offset: int = 0

        self.order_count: int = 1_000_000
        self.order_count_lock: Lock = Lock()
        self.connect_time: int = 0

    # 登录
    def sign(self, request: Request) -> Request:
        security: Security = request.data["security"]
        if security == Security.NONE:
            request.data = None
            return request

        if request.params:
            path: str = request.path + "?" + urllib.parse.urlencode(request.params)
        else:
            request.params = dict()
            path: str = request.path

        if security == Security.SIGNED:
            timestamp: int = int(time.time() * 1000)

            if self.time_offset > 0:
                timestamp -= abs(self.time_offset)
            elif self.time_offset < 0:
                timestamp += abs(self.time_offset)

            request.params["timestamp"] = timestamp

            query: str = urllib.parse.urlencode(sorted(request.params.items()))
            signature: bytes = hmac.new(self.secret, query.encode(
                "utf-8"), hashlib.sha256).hexdigest()

            query += "&signature={}".format(signature)
            path: str = request.path + "?" + query

        request.path = path
        request.params = {}
        request.data = {}

        headers = {
            "Content-Type": "application/x-www-form-urlencoded",
            "Accept": "application/json",
            "X-MBX-APIKEY": self.key,
            "Connection": "close"
        }

        if security in [Security.SIGNED, Security.API_KEY]:
            request.headers = headers
        return request

    # 连接 http api
    def connect(
            self,
            key: str,
            secret: str,
            proxy_host: str,
            proxy_port: int
    ) -> None:
        self.key = key
        self.secret = secret.encode()
        self.proxy_port = proxy_port
        self.proxy_host = proxy_host

        self.connect_time = (
                int(datetime.now().strftime("%y%m%d%H%M%S")) * self.order_count
        )
        self.init(F_REST_HOST, proxy_host, proxy_port)
        self.start()

        # 获取服务器时间
        self.query_time()
        # 保持用户 ws 的链接
        self.start_user_stream()

    # 查询服务器时间
    def query_time(self) -> None:
        data: dict = {
            "security": Security.NONE
        }

        path: str = "/dapi/v1/time"

        self.add_request(
            "GET",
            path,
            callback=self.on_query_time,
            on_failed=self.on_query_time_failed,
            on_error=self.on_query_time_error,
            data=data
        )

    # 自定义订单号
    def _new_order_id(self) -> int:
        """生成自定义订单ID"""
        with self.order_count_lock:
            self.order_count += 1
            return self.order_count

    def get_order_id(self) -> str:
        return "x-cLbi5uMH" + str(self.connect_time + self._new_order_id())

    # 发送订单
    def send_order(self, req: OrderRequest, orderid: str):
        order: OrderData = req.create_order_data(
            orderid,
        )
        self.gateway.on_order(order)

        data: dict = {
            "security": Security.SIGNED
        }
        if req.positionSide == PositionSide.ONEWAY:
            params: dict = {
                "symbol": req.symbol,
                "side": DIRECTION_VT2BINANCES[req.direction],
                "quantity": req.volume,
                "newClientOrderId": orderid,
                "newOrderRespType": "RESULT",
                "recvWindow": self.recv_window
            }
            if req.offset == Offset.CLOSE:
                params['reduceOnly'] = True
        else:
            if (req.direction == Direction.SHORT and req.offset == Offset.OPEN) or (
                    req.direction == Direction.LONG and req.offset == Offset.CLOSE):
                position_side = "SHORT"
            else:
                position_side = "LONG"

            params: dict = {
                "symbol": req.symbol,
                "side": DIRECTION_VT2BINANCES[req.direction],
                "positionSide": position_side,
                "quantity": req.volume,
                "newClientOrderId": orderid,
                "newOrderRespType": "RESULT",
                "recvWindow": self.recv_window
            }

        # 如果订单type = taker，那就是市价单，如果不是，则是限价单
        if req.type == OrderType.MARKET:
            params["type"] = "MARKET"  # 市价单
        elif req.type == OrderType.STOP_MARKET:
            params["type"] = "STOP_MARKET"  # 止损单
            params["stopPrice"] = req.stop_loss_price
        elif req.type == OrderType.TAKE_PROFIT_MARKET:
            params["type"] = "TAKE_PROFIT_MARKET"  # 止赢单
            params["stopPrice"] = req.stop_profit_price
        else:
            order_type, time_condition = ORDERTYPE_VT2BINANCES[req.type]
            params["type"] = order_type
            params["timeInForce"] = time_condition
            params["price"] = req.price

        path: str = "/dapi/v1/order"

        self.add_request(
            method="POST",
            path=path,
            callback=self.on_send_order,
            data=data,
            params=params,
            extra=order,
            on_error=self.on_send_order_error,
            on_failed=self.on_send_order_failed
        )

        return order.orderid

    # 撤销订单
    def cancel_order(self, req: CancelRequest) -> None:
        """cancel order"""
        data: dict = {
            "security": Security.SIGNED
        }

        params: dict = {
            "symbol": req.symbol,
            "origClientOrderId": req.orderid,
            "recvWindow": self.recv_window
        }

        path: str = "/dapi/v1/order"

        self.add_request(
            method="DELETE",
            path=path,
            callback=self.on_cancel_order,
            params=params,
            data=data,
            on_failed=self.on_cancel_order_failed,
        )

    # 获取历史K线数据
    def query_history(self, req: HistoryRequest) -> List[BarData]:
        history: List[BarData] = []
        limit: int = 1500
        end_time: int = int(datetime.timestamp(req.end))

        while True:
            # query parameters
            params: dict = {
                "symbol": req.symbol,
                "interval": INTERVAL_VT2BINANCES[req.interval],
                "limit": limit,
                "endTime": end_time * 1000
            }

            path: str = "/dapi/v1/klines"
            if req.start:
                start_time = int(datetime.timestamp(req.start))
                params["startTime"] = start_time * 1000  # convert the start time into milliseconds

            resp: Response = self.request(
                "GET",
                path=path,
                data={"security": Security.NONE},
                params=params
            )

            # will break the while loop if the request failed
            if resp.status_code // 100 != 2:
                msg: str = f"查询历史kline数据失败, status code: {resp.status_code}, msg: {resp.text}"
                print(msg)
                break
            else:
                data: dict = resp.json()
                if not data:
                    msg: str = f"历史K线数据为空，开始时间: {start_time}"
                    print(msg)
                    break

                buf: List[BarData] = []

                for row in data:
                    bar: BarData = BarData(
                        symbol=self.gateway.symbolMap[req.symbol],
                        datetime=generate_datetime(row[0]),
                        endTime=generate_datetime(row[6]),
                        interval=req.interval,
                        volume=float(row[5]),
                        turnover=float(row[7]),
                        open_price=float(row[1]),
                        high_price=float(row[2]),
                        low_price=float(row[3]),
                        close_price=float(row[4]),
                        exchange=self.gateway.exchange
                    )
                    buf.append(bar)

                begin: datetime = buf[0].datetime
                end: datetime = buf[-1].datetime

                buf = list(reversed(buf))
                history.extend(buf)
                msg: str = f"查询历史kline数据成功, {req.symbol} - {req.interval.value}，{begin} - {end}"
                print(msg)

                # if the data len is less than limit, break the while loop
                if len(data) < limit:
                    break

                # update start time
                end_dt = begin - TIMEDELTA_MAP[req.interval]
                end_time = int(datetime.timestamp(end_dt))

        history = list(reversed(history))
        return history

    # 开始ws 交易数据推送
    def start_user_stream(self) -> None:
        """获取ws listen key 帮助重新连接ws """
        data: dict = {
            "security": Security.API_KEY
        }

        path: str = "/dapi/v1/listenKey"

        self.add_request(
            method="POST",
            path=path,
            callback=self.on_start_user_stream,
            on_failed=self.on_start_user_stream_failed,
            on_error=self.on_start_user_stream_error,
            data=data
        )

    # 撤销订单成功回调
    def on_cancel_order(self, data: dict, request: Request) -> None:
        key: Tuple[str, str] = (data.get("type"), data.get("timeInForce"))
        order_type: OrderType = ORDERTYPE_BINANCES2VT.get(key, OrderType.LIMIT)

        traded = float(data.get("executedQty", "0"))
        traded_price = float(data.get('avgPrice', '0'))

        price = float(data["price"])
        if price <= 0 < traded_price:
            price = traded_price

        order: OrderData = OrderData(
            orderid=data.get("clientOrderId"),
            symbol=data.get("symbol"),
            price=price,
            volume=float(data.get("origQty")),
            traded=traded,
            traded_price=traded_price,
            type=order_type,
            direction=DIRECTION_BINANCES2VT.get(data.get("side")),
            status=STATUS_BINANCES2VT.get(data.get("status"), Status.CANCELLED),
            datetime=generate_datetime(float(data.get("updateTime", time.time() * 1000))),
            exchange=self.gateway.exchange
        )
        self.gateway.on_order(order)

    # 撤销订单失败回调
    def on_cancel_order_failed(self, status_code: int, request: Request) -> None:
        self.failed_with_timestamp(request)
        orderid = ""
        if request.extra:
            order: OrderData = copy(request.extra)
            orderid = order.orderid
            order.status = Status.REJECTED
            self.gateway.on_order(copy(order))

        msg = f"撤销订单失败, orderid: {orderid}, status code：{status_code}, msg：{request.response.text}"
        print(msg)

    # 延长ws 数据保持时间
    def keep_user_stream(self) -> None:
        """延长 listenKey 过期时间"""
        self.keep_alive_count += 1
        if self.keep_alive_count < 300:
            return None

        self.keep_alive_count = 0

        data: dict = {
            "security": Security.API_KEY
        }

        params: dict = {
            "listenKey": self.user_stream_key
        }

        path: str = "/dapi/v1/listenKey"

        self.add_request(
            method="PUT",
            path=path,
            callback=self.on_keep_user_stream,
            params=params,
            data=data,
            on_failed=self.on_keep_user_stream_failed,
            on_error=self.on_keep_user_stream_error
        )

    # 查询服务器时间成功回调
    def on_query_time(self, data: dict, request: Request) -> None:
        """查询服务器时间回调"""
        local_time: int = int(time.time() * 1000)
        server_time: int = int(data["serverTime"])
        self.time_offset: int = local_time - server_time

    # 查询服务器时间失败回调
    def on_query_time_failed(self, status_code: int, request: Request):
        self.query_time()

    # 查询服务器时间错误回调
    def on_query_time_error(self, exception_type: type, exception_value: Exception, tb, request: Request) -> None:
        self.query_time()

    # 发送订单成功回调
    def on_send_order(self, data: dict, request: Request) -> None:
        pass

    # 发送订单失败回调
    def on_send_order_failed(self, status_code: int, request: Request) -> None:
        self.failed_with_timestamp(request)
        if request.extra:
            order: OrderData = copy(request.extra)
            order.status = Status.REJECTED
            order.rejected_reason = request.response.text if request.response.text else ""
            self.gateway.on_order(order)

            msg: str = f"send order failed, orderid: {order.orderid}, status code：{status_code}, msg：{request.response.text}"
            print(msg)

    # 发送订单错误回调
    def on_send_order_error(
            self, exception_type: type, exception_value: Exception, tb, request: Request
    ) -> None:
        if request.extra:
            order: OrderData = copy(request.extra)
            order.status = Status.REJECTED
            order.rejected_reason = "on_send_order_error"
            self.gateway.on_order(order)

        if not issubclass(exception_type, (ConnectionError, SSLError)):
            self.on_error(exception_type, exception_value, tb, request)

    # 开始获取ws stream 回调
    def on_start_user_stream(self, data: dict, request: Request) -> None:
        """查询listenkey回调，然后连接trade ws"""
        # 成功获取，记录该参数
        self.user_stream_key = data["listenKey"]
        self.keep_alive_count = 0

        url = F_WEBSOCKET_TRADE_HOST + self.user_stream_key
        self.trade_ws_api.connect(url, self.proxy_host, self.proxy_port)

    # 开始获取ws stream 失败处理
    def on_start_user_stream_failed(self, status_code: int, request: Request):
        self.failed_with_timestamp(request)  # 重新获取服务器时间撮
        self.start_user_stream()  # 再次链接

    # 开始获取ws stream 错误处理
    def on_start_user_stream_error(self, exception_type: type, exception_value: Exception, tb, request: Request):
        self.start_user_stream()  # 再次链接

    # 延长监听密钥过期时间请求回调
    def on_keep_user_stream(self, data: dict, request: Request) -> None:
        self.keep_alive_failed_count = 0

    # 延长监听密钥过期失败回处理
    def on_keep_user_stream_failed(self, status_code: int, request: Request):
        self.failed_with_timestamp(request)
        self.keep_alive_failed_count += 1
        if self.keep_alive_failed_count <= 3:
            self.keep_alive_count = 1200000
            self.keep_user_stream()
        else:
            self.keep_alive_failed_count = 0
            self.start_user_stream()

    # 延长监听密钥过期错误回调处理
    def on_keep_user_stream_error(
            self, exception_type: type, exception_value: Exception, tb, request: Request
    ) -> None:
        self.keep_alive_failed_count += 1
        if self.keep_alive_failed_count <= 3:
            self.keep_alive_count = 1200000
            self.keep_user_stream()
        else:
            self.keep_alive_failed_count = 0
            self.start_user_stream()

        if not issubclass(exception_type, TimeoutError):
            self.on_error(exception_type, exception_value, tb, request)

    # 因为时间撮失败
    def failed_with_timestamp(self, request: Request):
        try:
            if request and request.response and request.response.text:
                resp = json.loads(request.response.text)
                if resp.get('code') == -1021:
                    self.query_time()
        except Exception:
            pass


class BinanceInverseTradeWebsocketApi(WebsocketClient):

    def __init__(self, gateway: BinanceFutureInverseGateway) -> None:
        super().__init__()

        self.gateway: BinanceFutureInverseGateway = gateway

    def connect(self, url: str, proxy_host: str, proxy_port: int) -> None:
        """连接 binance usdt/busd 期货交易 ws"""
        self.init(url, proxy_host, proxy_port)
        self.start()

    def on_connected(self) -> None:
        pass

    def on_packet(self, packet: dict) -> None:
        """从 ws 接收数据s"""
        if packet["e"] == "ORDER_TRADE_UPDATE":
            self.on_order(packet)

    def on_exit_loop(self):
        """ws 连接环"""
        self.gateway.rest_api.start_user_stream()

    # 订单ws 推送
    def on_order(self, packet: dict) -> None:
        ord_data: dict = packet["o"]
        key: Tuple[str, str] = (ord_data["o"], ord_data["f"])
        order_type: OrderType = ORDERTYPE_BINANCES2VT.get(key, OrderType.LIMIT)
        price = float(ord_data["p"])
        if price <= 0:
            price = float(ord_data['ap'])

        order: OrderData = OrderData(
            symbol=ord_data["s"],
            orderid=str(ord_data["c"]),
            type=order_type,
            direction=DIRECTION_BINANCES2VT[ord_data["S"]],
            price=price,
            volume=float(ord_data["q"]),
            traded=float(ord_data["z"]),
            traded_price=float(ord_data.get("L", "0")),
            status=STATUS_BINANCES2VT.get(ord_data["X"], Status.NOTTRADED),
            datetime=generate_datetime(packet["E"]),
            exchange=self.gateway.exchange
        )
        self.gateway.on_order(copy(order))


class BinanceInverseDataWebsocketApi(WebsocketClient):

    def __init__(self, gateway: BinanceFutureInverseGateway) -> None:
        """"""
        super().__init__()

        self.gateway: BinanceFutureInverseGateway = gateway
        self.ticks: Dict[str, TickData] = {}
        self.bars: Dict[str, BarData] = {}
        self.reqid: int = 0

    def connect(
            self,
            proxy_host: str,
            proxy_port: int,
    ) -> None:
        """连接ws数据"""
        self.init(F_WEBSOCKET_DATA_HOST, proxy_host, proxy_port)
        self.start()

    def on_connected(self) -> None:
        if self.ticks:
            channels = []
            for symbol in self.ticks.keys():
                channels.append(f"{symbol}@aggTrade")

            req: dict = {
                "method": "SUBSCRIBE",
                "params": channels,
                "id": self.reqid
            }
            self.send_packet(req)

    def subscribe(self, symbol: str, interval: Interval = None) -> None:
        """订阅数据"""
        # 如果间隔参数不为 None, 那么就是订阅K线数据
        if interval:
            # 订阅 K线 流程
            if symbol in self.bars:
                return
            self.reqid += 1
            bar: BarData = BarData(
                symbol=symbol,
                datetime=datetime.now(),
                endTime=datetime.now(),
                interval=interval,
                exchange=self.gateway.exchange,
                product=self.gateway.product
            )
            self.bars[symbol.lower() + f"{interval.value}"] = bar
            channels = [
                f"{symbol.lower()}@kline_{interval.value}"
            ]
        else:
            # 订阅tick 数据流程
            if symbol in self.ticks:
                return
            self.reqid += 1
            tick: TickData = TickData(
                symbol=symbol,
                datetime=datetime.now(),
                exchange=self.gateway.exchange,
                product=self.gateway.product
            )
            self.ticks[symbol.lower()] = tick
            channels = [
                f"{symbol.lower()}@aggTrade"
            ]

        req: dict = {
            "method": "SUBSCRIBE",
            "params": channels,
            "id": self.reqid
        }
        self.send_packet(req)

    def on_packet(self, packet: dict):
        stream: str = packet.get("stream", None)
        if not stream:
            return

        data: dict = packet["data"]

        symbol, channel = stream.split("@")
        bar: BarData = self.bars.get(symbol + f"{channel.split('_')[-1]}")
        tick: TickData = self.ticks.get(symbol)

        if channel.startswith("kline_"):
            data = data['k']
            if data['x']:
                bar.volume = float(data['v'])
                bar.turnover = float(data['q'])
                bar.open_price = float(data['o'])
                bar.high_price = float(data['h'])
                bar.low_price = float(data['l'])
                bar.close_price = float(data['c'])
                bar.datetime = generate_datetime(float(data['t']))
                bar.endTime = generate_datetime(float(data['T']))
                self.gateway.on_bar(copy(bar))

        elif channel == "aggTrade":
            # tick.volume = float(data['v'])
            # tick.turnover = float(data['q'])
            # tick.open_price = float(data['o'])
            # tick.high_price = float(data['h'])
            # tick.low_price = float(data['l'])
            tick.last_price = float(data['p'])
            tick.datetime = generate_datetime(float(data['T']))

            if tick.last_price:
                tick.localtime = datetime.now()
                self.gateway.on_tick(copy(tick))


def generate_datetime(timestamp: float) -> datetime:
    """generate time"""
    dt: datetime = datetime.fromtimestamp(timestamp / 1000)
    return dt



# 导入 strategy_api 包里面的策略模版
from strategy_api.strategies.template import StrategyTemplate
# 导入 BINANCE U 本位合约 API包
from strategy_api.tm_api.Okex.futureUsdt import OkexFutureUsdtGateway
from strategy_api.tm_api.Binance.futureUsdt import BinanceFutureUsdtGateway
from strategy_api.tm_api.Binance.futureInverse import BinanceFutureInverseGateway

# 导入 strategy_api 中使用的常量对象
from strategy_api.tm_api.object import Interval, BarData, OrderData, Status, PositionSide, DataType


# 策略类
class StrategyDemo(StrategyTemplate):
    # 属性 作者(标志该策略的开发人员)
    author = "DYX"

    # 初始化方法
    def __init__(self):
        super(StrategyDemo, self).__init__()

    # 初始化策略参数
    def init_parameters(self):
        self.buy_switch = True
        self.long_id = ""
        self.stop_loss_id = ""
        self.stop_profit_id = ""
        self.volume = 1
        self.rate_stop = 0.008

    # k 线数据的回调, 可以在该方法里面记录 k 线数据、分析k线数据
    def on_bar(self, bar: BarData):
        okex_api = self.get_gateway("okex_api")
        # print(bar)
        if self.buy_switch:
            # 开多单
            self.long_id = okex_api.new_order_id()

            okex_api.buy(
                orderid=self.long_id,
                symbol="SOL-USD-SWAP",
                volume=self.volume,
                price=10
            )
            print("--------------做多")
            self.buy_switch = False

    # 获取历史k线，获取最新一根k线的开盘价
    def query_history_kline(self):
        okex_api = self.get_gateway("okex_api")
        try:
            kls = okex_api.query_history(symbol="SOL-USD-SWAP", minutes=5, interval=Interval.MINUTE)
        except Exception as e:
            kls = okex_api.query_history(symbol="SOL-USD-SWAP", minutes=5, interval=Interval.MINUTE)

        open_price = kls[-1].open_price
        return open_price

    # 订单 数据的回调，订单状态的改变都会通过websoket 推送到这里，例如 从提交状态 改为 全成交状态，或者提交状态 改为 撤销状态 都会推送
    # 可以在这里对仓位进行一个记录
    def on_order(self, order: OrderData):
        okex_api = self.get_gateway("okex_api")
        if order.status == Status.ALLTRADED and self.long_id == order.orderid:
            print("做多成交")
            open_price = self.query_history_kline()
            print(f"最新开盘价：{open_price}")
            self.stop_profit_id = okex_api.new_order_id()
            okex_api.sell(
                orderid=self.stop_profit_id,
                symbol="SOL-USD-SWAP",
                volume=self.volume,
                price=round(open_price * (1 + self.rate_stop), 3),
                stop_profit=True
            )

            self.stop_loss_id = okex_api.new_order_id()
            okex_api.sell(
                orderid=self.stop_loss_id,
                symbol="SOL-USD-SWAP",
                volume=self.volume,
                price=round(open_price * (1 - self.rate_stop), 3),
                stop_loss=True
            )

        elif order.status == Status.ALLTRADED and (
                self.stop_profit_id == order.orderid or self.stop_loss_id == order.orderid):
            # okex 不用测小止损止盈单，只要止损了，止盈单自动撤销

            if self.stop_loss_id == order.orderid:
                print("止损")
                pass
                # okex_api.cancel_order(orderid=self.stop_profit_id, symbol="SOL-USD-SWAP")
            elif self.stop_profit_id == order.orderid:
                print("止盈")
                pass
                # okex_api.cancel_order(orderid=self.stop_loss_id, symbol="SOL-USD-SWAP")

            self.buy_switch = True


def start_strategy(api_setting):
    # 初始化策略
    s = StrategyDemo()

    # 添加 BINANCE U本位网关
    okex_gateway = s.add_gateway(BinanceFutureInverseGateway, "okex_api", api_setting)

    # 订阅数据
    okex_gateway.subscribe(symbol="BTC-USD-PERP", data_type=DataType.BAR, interval=Interval.MINUTE)


if __name__ == '__main__':
    print("启动量化系统: 等待策略运行")
    api_setting = {
        "key": "",
        "secret": "",
        "proxy_host": "",
        "proxy_port": 0,
        "Passphrase": "",

        "call_order_switch": False,  # 订单数据回调 开关
        "call_tick_switch": False,  # tick数据回调 开关
        "call_bar_switch": False  # k线数据回调 开关
    }
    start_strategy(api_setting)



import time

from strategy_api.event.engine import Event, EventEngine, EVENT_TICK, EVENT_BAR, EVENT_ORDER
from strategy_api.tm_api.Binance.futureUsdt import BinanceFutureUsdtGateway
from strategy_api.tm_api.Binance.futureInverse import BinanceFutureInverseGateway
from strategy_api.tm_api.Okex.futureUsdt import OkexFutureUsdtGateway

from strategy_api.tm_api.object import Interval, TickData, BarData, OrderData, DataType, PositionSide

api_setting = {
    "key": "",
    "secret": "",
    "proxy_host": "",
    "proxy_port": 0,
    "Passphrase": ""
}


def process_tick_event(event: Event):
    tick: TickData = event.data
    print("tick 数据：")
    print(tick)


def process_bar_event(event: Event):
    bar: BarData = event.data
    print("k 线 数据：")
    print(bar)


def process_order_event(event: Event):
    order: OrderData = event.data
    print("订单 数据：")
    print(order)

# # ---------------------------币安api---------------------------


# event_engine = EventEngine()
# event_engine.start()
# event_engine.register(EVENT_TICK, process_tick_event)
# event_engine.register(EVENT_BAR, process_bar_event)
# event_engine.register(EVENT_ORDER, process_order_event)
#
# # api = BinanceFutureUsdtGateway(event_engine)
# api = BinanceFutureInverseGateway(event_engine)
# # # 链接api
# api.connect(api_setting)
# time.sleep(5)
# history_kline = api.query_history(symbol="BTCUSD_PERP", interval=Interval.MINUTE, minutes=20)
# print(len(history_kline))
# #
# # 订阅 1 分钟 K 线 行情
# api.subscribe(symbol="BTCUSD_PERP", data_type=DataType.BAR, interval=Interval.MINUTE)

# 订阅 tick 行情
# api.subscribe(symbol="BTCUSDT", data_type=DataType.TICK)

# 下单
# api.buy(symbol="BTCUSDT", volume=0.01, price=9999, maker=True, stop_loss=False, stop_profit=False)
#
# # 撤销订单
# api.cancel_order(orderid="xl_1111111", symbol="BTCUSDT")
# # -------------------------------------------------------------
# ---------------------------Okex api---------------------------


# 订阅 1 分钟 K 线 行情
# api.subscribe(symbol="BTC-USDT", interval=Interval.MINUTE)

# # 订阅 tick 行情



event_engine = EventEngine()
event_engine.start()
event_engine.register(EVENT_TICK, process_tick_event)
event_engine.register(EVENT_BAR, process_bar_event)
event_engine.register(EVENT_ORDER, process_order_event)

api = OkexFutureUsdtGateway(event_engine)

# 链接api
api.connect(api_setting)
time.sleep(5)
# 查询历史K线数据
# history_kline = api.query_history(symbol="BTC-USDT-SWAP", interval=Interval.MINUTE, minutes=20)
# for l in history_kline:
#     print(l)
# api.subscribe(symbol="BTC-USDT-SWAP", data_type=DataType.BAR, interval=Interval.MINUTE)

# api.subscribe(symbol="BTC-USDT-SWAP", data_type=DataType.TICK)
# 市价单下单
# api.short(symbol="SOL-USDT-SWAP", volume=1, price=9999, maker=False, stop_loss=False, stop_profit=False, position_side=PositionSide.TWOWAY)
# 限价单下单
api.buy(symbol="SOL-USD-SWAP", volume=1, price=10, maker=False, stop_loss=False, stop_profit=False)
# print(orderid)
# # 撤销订单
# api.cancel_order(orderid="230104211950000001", symbol="SOL-USDT-SWAP")
# -------------------------------------------------------------




import datetime
import time
from abc import ABC
from typing import List, Dict, Type
from strategy_api.event.engine import EventEngine, Event, EVENT_TICK, EVENT_BAR, EVENT_ORDER
from strategy_api.tm_api.base import BaseGateway
from strategy_api.tm_api.object import TickData, BarData, OrderData, Interval


class StrategyTemplate(ABC):
    author: str = ""

    def __init__(self) -> None:
        self.event_engine: EventEngine = EventEngine()
        self.gateways: Dict[str, BaseGateway] = {}
        self.register_handler(self.process_tick_event, self.process_bar_event, self.process_order_event)

        self.init_strategy()    # 初始化引擎

    def get_gateway(self, gateway_name) -> BaseGateway:
        gateway: BaseGateway = self.gateways.get(gateway_name)
        if not gateway:
            print(f"没有该网关: {gateway_name}")
        return gateway

    def init_strategy(self):
        # 启动事件引擎
        self.event_engine.start()
        # 初始化参数
        self.init_parameters()

    # 添加网关
    def add_gateway(self,
                    gateway_class: Type[BaseGateway],
                    gateway_name,
                    api_setting: dict
                    ) -> BaseGateway:
        order_call_switch = api_setting.get("call_order_switch", False)
        bar_call_switch = api_setting.get("call_bar_switch", False)
        tick_call_switch = api_setting.get("call_tick_switch", False)

        gateway: BaseGateway = gateway_class(self.event_engine, order_call_switch, bar_call_switch, tick_call_switch)
        self.gateways[gateway_name] = gateway
        gateway.connect(api_setting)
        print(f"添加网关:{gateway_name}, 等待网关链接")
        time.sleep(5)
        return gateway

    # 事件注册
    def register_handler(self, tick_handler: callable, bar_handler: callable, order_handler: callable):
        self.event_engine.register(EVENT_TICK, tick_handler)
        self.event_engine.register(EVENT_BAR, bar_handler)
        self.event_engine.register(EVENT_ORDER, order_handler)

    # tick 数据事件推送
    def process_tick_event(self, event: Event):
        tick: TickData = event.data
        self.on_tick(tick)

    # k线 数据推送
    def process_bar_event(self, event: Event):
        bar: BarData = event.data
        self.on_bar(bar)

    # 订单 数据推送
    def process_order_event(self, event: Event):
        order: OrderData = event.data
        self.on_order(order)

    # 策略参数设置
    def init_parameters(self):
        pass

    # 策略启动时的回调。
    def on_start(self) -> None:
        pass

    # 新的tick数据更新回调。
    def on_tick(self, tick: TickData) -> None:
        pass

    # 新的bar数据更新回调。
    def on_bar(self, bar: BarData) -> None:
        pass

    # 新订单数据更新回调。
    def on_order(self, order: OrderData) -> None:
        pass

    # 记录数据
    def record_bar(self, bar: BarData):
        pass

    # 数据分析
    def deal_data(self, bar: BarData):
        pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# @Author   : Kevin
# @Time     : 2022/12/12

import setuptools

with open('README.md', 'r', encoding='utf-8') as fh:
    long_description = fh.read()

setuptools.setup(
    name='strategy_api',
    version='1.2.8.1',
    author='Kevin',
    author_email='1782552261@qq.com',
    description='bug 修改',
    long_description=long_description,
    url='https://gitee.com/Jason520deng/strategy_api',
    packages=setuptools.find_packages(),
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
)




import json
import sys
import traceback
from datetime import datetime
from types import coroutine
from threading import Thread
from asyncio import (
    get_event_loop,
    set_event_loop,
    run_coroutine_threadsafe,
    AbstractEventLoop
)

from aiohttp import ClientSession, ClientWebSocketResponse


class WebsocketClient:

    def __init__(self):
        self._active: bool = False
        self._host: str = ""

        self._session: ClientSession = ClientSession()
        self._ws: ClientWebSocketResponse = None
        self._loop: AbstractEventLoop = None

        self._proxy: str = ""
        self._ping_interval: int = 60  # ping interval for 60 seconds
        self._header: dict = {}

        self._last_sent_text: str = ""
        self._last_received_text: str = ""

    def init(
            self,
            host: str,
            proxy_host: str = "",
            proxy_port: int = 0,
            ping_interval: int = 60,
            header: dict = None
    ):
        self._host = host
        self._ping_interval = ping_interval

        if header:
            self._header = header

        if proxy_host and proxy_port:
            self._proxy = f"http://{proxy_host}:{proxy_port}"

    def start(self):
        try:
            if self._ws:
                coro = self._ws.close()
                run_coroutine_threadsafe(coro, self._loop)
        except Exception as error:
            pass

        if self._active:
            return None

        self._active = True

        if not self._loop:
            self._loop = get_event_loop()
        start_event_loop(self._loop)

        run_coroutine_threadsafe(self._run(), self._loop)

    def stop(self):
        self._active = False

        if self._ws:
            coro = self._ws.close()
            run_coroutine_threadsafe(coro, self._loop)
            print("websoket 停止")

        if self._loop and self._loop.is_running():
            self._loop.stop()
            print("websoket 停止")

    def join(self):
        pass

    def send_packet(self, packet: dict):
        if self._ws:
            text: str = json.dumps(packet)
            self._record_last_sent_text(text)

            coro: coroutine = self._ws.send_str(text)
            run_coroutine_threadsafe(coro, self._loop)

    def unpack_data(self, data: str):
        return json.loads(data)

    def on_connected(self):
        pass

    def on_disconnected(self):
        pass

    def on_packet(self, packet: dict):
        pass

    def on_error(
            self,
            exception_type: type,
            exception_value: Exception,
            tb
    ) -> None:
        try:
            print("WebsocketClient on error" + "-" * 10)
            print(self.exception_detail(exception_type, exception_value, tb))
        except Exception:
            traceback.print_exc()

    def on_exit_loop(self):
        self.start()

    def exception_detail(
            self,
            exception_type: type,
            exception_value: Exception,
            tb
    ) -> str:
        text = "[{}]: Unhandled WebSocket Error:{}\n".format(
            datetime.now().isoformat(), exception_type
        )
        text += "LastSentText:\n{}\n".format(self._last_sent_text)
        text += "LastReceivedText:\n{}\n".format(self._last_received_text)
        text += "Exception trace: \n"
        text += "".join(
            traceback.format_exception(exception_type, exception_value, tb)
        )
        return text

    async def _run(self):
        while self._active:
            # try catch error/exception
            try:
                # connect ws server
                if self._proxy:
                    self._ws = await self._session.ws_connect(
                        self._host,
                        proxy=self._proxy,
                        verify_ssl=False
                    )
                else:
                    self._ws = await self._session.ws_connect(
                        self._host
                    )
                # call the on_connected function
                self.on_connected()

                # receive data from websocket
                async for msg in self._ws:
                    text: str = msg.data
                    self._record_last_received_text(text)

                    data: dict = self.unpack_data(text)
                    self.on_packet(data)

                # remove the _ws object
                self._ws = None

                # call the on_disconnected
                self.on_disconnected()
            # on exception
            except Exception:
                et, ev, tb = sys.exc_info()
                self.on_error(et, ev, tb)
                break

        self._active = False
        self.on_exit_loop()

    def _record_last_sent_text(self, text: str):
        self._last_sent_text = text[:1000]

    def _record_last_received_text(self, text: str):
        self._last_received_text = text[:1000]


def start_event_loop(loop: AbstractEventLoop) -> AbstractEventLoop:
    if not loop.is_running():
        thread = Thread(target=run_event_loop, args=(loop,))
        thread.daemon = True
        thread.start()


def run_event_loop(loop: AbstractEventLoop) -> None:
    set_event_loop(loop)
    loop.run_forever()
